/*
 * "Bistro" firmware  
 * This code runs on a single-channel or dual-channel driver (FET+7135)
 * with an attiny25/45/85 MCU and a capacitor to measure offtime (OTC).
 *
 * Original version Copyright (C) 2015 Selene Scriven, 
 * Modified significantly by Texas Ace (triple mode groups) and Flintrock (code size, Vcc, OTSM, eswitch, 4-chan, delay-sleep, etc.) 
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 *
 *  TA adds many great custom mode groups and nice build scripts
 *  Flintrock Updates:
 *    Implements OTSM "off-time sleep mode" to determine click length based on watchdog timer with mcu powered by CAP.
 *    Adds Vcc inverted voltage read option, with calibration table, allows voltage read of 1S lights without R1/R2 divider hardware.
 *    Implemented 4th PWM channel. (untested, it could even work!)
 *    Now supports E-switches.
 *    Algorithm-generated calibration tables now allow voltage adjustment with one or two configs in fr-calibration.h
 *    Modified tk-attiny.h to simplify layout definitions and make way for more easily-customizable layouts.     
 *    Added BODS to voltage shutdown (and E-switch poweroff) to protect low batteries longer (~4uA for supported chips).
 *    Implements space savings (required for other features):
 *    ... See Changes.txt for details.
 *
 *    see fr-tk-attiny.h for mcu layout.
 *
 * FUSES
 *
 * ##fuses (for attin25/45/95): high: 
 *            DD => BOD disabled early revision chips may not work with BOD and OTSM, but disabling it risks corruption.
 *			  DE => BOD set to 1.8V  Non V-version chips are still not specced for this and may corrupt.  late model V chips with this setting is best.
 *            DF => BOD set to 2.7V  Safe on all chips, but may not work well with OTSM, without huge caps.
 *
 *       low  C2, 0ms startup time.  Probably won't work with BODS capable late model chips, but should improve OTSM a little if not using one anyway. 
 *            D2, 4ms  Testing finds this seems to work well on attiny25 BODS capable chip, datasheet maybe implies 64 is safer (not clear and hardware dependent). 
              E2  64ms startup.  Will probably consume more power during off clicks and might break click timing.
 *      
 *        Ext: 0xff
 *       Tested with high:DE low:D2, Ext 0xff on non-V attiny25, but it's probably not spec compliant with that chip.
 *
 *      For more details on these settings
 *      http://www.engbedded.com/cgi-bin/fcx.cgi?P_PREV=ATtiny25&P=ATtiny25&M_LOW_0x3F=0x12&M_HIGH_0x07=0x06&M_HIGH_0x20=0x00&B_SPIEN=P&B_SUT0=P&B_CKSEL3=P&B_CKSEL2=P&B_CKSEL0=P&B_BODLEVEL0=P&V_LOW=E2&V_HIGH=DE&V_EXTENDED=FF
 *
 * STARS
 *      Star 3 = unused
 *
 * CALIBRATION
 *   FR's new method Just adjust parameters in fr-calibration.h
 *   Now uses something between fully caluclated and fully measured method.
 *   You can make simple adjustments to the calculations to get a good result.
 *
 *   To find out what values to use, flash the driver with battcheck.hex
 *   and hook the light up to each voltage you need a value for.  This is
 *   much more reliable than attempting to calculate the values from a
 *   theoretical formula.
 *
 *   Same for off-time capacitor values.  Measure, don't guess.
 *
 * 
 */
// Won't compile with -O0 (sometimes useful for debugging) without following:
#ifndef __OPTIMIZE__
#define inline
#endif

#include <avr/pgmspace.h>
//#include <avr/io.h>
#include <avr/interrupt.h>
#include <avr/eeprom.h>
#include <avr/sleep.h>
#include <avr/power.h>
#include <util/delay_basic.h>
//#include <string.h>

/*
 * =========================================================================
 * Settings to modify per driver
 */

// Choose your MCU here, or in the build script
//  choices are now 13, 25, 45, and 85.  Yes, 45 and 85 are now different

//#define ATTINY 13
#define ATTINY 25
//#define ATTINY 45  // yes these are different now, hopefuly only in ways we already know.
//#define ATTINY 85  //   bust specifically they have a 2 byte stack pointer that OTSM accesses.

/////////////// Choose a layout////////////////

//#define FET_7135_LAYOUT  // specify an I/O pin layout
#define TRIPLEDOWN_LAYOUT  // specify an I/O pin layout
//#define QUADRUPLEDOWN_LAYOUT                     // Yes, that right :)
//#define CUSTOM_LAYOUT  // go define it for however you want to wire your driver.

//**** You can now easily customize layouts in fr-tk-attiny.h :

#include "fr-tk-attiny.h" // should be compatible with old ones, but it might not be, so renamed it.

/*************************  Misc utility configs*************************************/
//#define VOLTAGE_CAL  // builds a driver that just blinks ADC reading, set the VOLTAGE mode below.
//#define STRIPPED // use for debugging, strips many features to make more space for debug compiler options or testing functions.
                   // customize it below
//#define OTSM_debug //may require STRIPPED to fit. //Will blink out number of wakes (see SLEEP_TIME_EXPONENT) on short clicks, before entering mode.

/************************************OFF-TIME CONFIGS**********************************/

#define OFFTIM3      // Use short/med/long off-time presses instead of just short/long


//#define USE_OTC // use off time cap. 

// Off-time sleep mode by -FR, thanks to flashy Mike and Mike C for early feasibility test with similar methods.
#define USE_OTSM  // USE OTSM.  Pin must be defined in the layout too.

//#define OTSM_USES_OTC // use OTC cap for extra power on OTSM (sets it output high to charge up)
#define OTSM_powersave // Squeeze out a bit more off-time by saving power during
             // shutoff detection (so at all times).  Implements ms resolution (could be less) idle sleeps in place of delay.
             // Seems to add at 0.5s of sleep at 3.1V 30uF cap, starting with only 0.75 that matters.  
			 // But at higher voltages, it's still only 0.5s additional, not multiplicative.
             // Also adds about 12 bytes that could be used for about two more mode groups instead :P


// Traditionally Cap values defined in tk-calibration
// But wake times in OTSM are not a calibration, just a configuration, so they go here.
// Times are in decimal seconds.  short (med) wakes are <, not = to wake_time_short(med)
#define SLEEP_TIME_EXPONENT 4   // OTSM clicky sleep will be increments of 16ms*2^STE  
                               // so 0 is 16, 1 is 32, 2 is 64, 3 is 128, 4 is 0.25s, 5 is 0.5s etc.
// To allow long click times use 4. For 1/8s resolution, use 3.
#define wake_time_short 0.5   // 0.5s : Short press is up to 0.5 s
#define wake_time_med   1.25   // this is limited by cap performance, but setting it high won't hurt 
                              // Since anything beyond the cap capacity will be read as long anyway.
							  // It's also the long-press off-threshold for e-switch operation.

/*****************************END OTSM CONGIGS*********************************/

//#define USE_ESWITCH  // pin must be defined in the layout too.

/**********************VOLTAGE CONFIG****************************************/

#define VOLTAGE_MON         // Comment out to disable LVP
//You should leave one (but only one) of the next two uncommented.  
#define READ_VOLTAGE_FROM_VCC  // inverted "internal" Vcc voltage monitoring
                               // Works well for 1S lights without worrying about resistor values.
//#define READ_VOLTAGE_FROM_DIVIDER  // classic voltage reading
#define REFERENCE_DIVIDER_READS_TO_VCC // default is 1.1V, but this is needed for divider reading with OTSM on the voltage pin.
                                         // This should normally be used with an LDO.  For 1S (non-LDO or 5.0VLDO) just avoid the problem with READ_VOLTAGE_FROM_VCC.


/*** Enable battery indicator mode?   */
#define USE_BATTCHECK
// Choose a battery indicator style
//#define BATTCHECK_4bars  // up to 4 blinks
//#define BATTCHECK_8bars  // up to 8 blinks
#define BATTCHECK_VpT  // Volts + tenths

/******theremal protection:  ***/
#define TEMPERATURE_MON          // You can set starting temperature in the "maxtemp" setting in config options first boot options.




// Options for first bootup/default:

#define USE_FIRSTBOOT // FR notes this only costs two bytes since not using it implements alternative checks anyway.

#define INIT_MODEGROUP     11       // which mode group will be default, mode groups below start at zero, select the mode group you want and subtract one from the number to get it by defualt here
#define INIT_ENABLE_MOON    1       // Should we add moon to the set of modes?
#define INIT_REVERSE_MODES  0       // flip the mode order?
#define INIT_MEMORY         1       // mode memory, or not
#define INIT_OFFTIM3        1       // enable medium-press by default?
#define INIT_MUGGLE_MODE    0       // simple mode designed for mugglesotsm
#define INIT_LOCKSWITCH     0       // E-swtich enabled.
#define INIT_MAXTEMP       88       // maximum temperature

#define BLINK_SPEED         750


// This is used to simplify the toggle function
// eliminating mode_override, using this threshold instead:
#define MINIMUM_OVERRIDE_MODE 245  // DO NOT EDIT.  DO NOT DEFINE ANY STROBES HIGER OR EQUAL TO THIS.

/******* Select Which Strobes to compile in**** *********************/
////Must still be connected to a mode group or hidden mode to be used, they won't get optimized out the code if unselected.
//Some other non-user configurable defines and globals.

#define BATTCHECK        244      // Convenience code for battery check mode
//mode codes for strobes, must be less than MINIMUM_OVERRIDE_MODE
#define BIKING_STROBE    243     // Single flash biking strobe mode
//#define FULL_BIKING_STROBE     // Stutter bike strobe, uncomment to enable
//#define POLICE_STROBE    242     // Dual mode alternating strobe
//#define RANDOM_STROBE    241
//#define SOS              240
//#define STROBE_8HZ       239
#define STROBE_10HZ      238
#define STROBE_16HZ      237
#define STROBE_OLD_MOVIE 236
#define STROBE_CREEPY    235     // Creepy strobe mode, or really cool if you have a bunch of friends around
//#define RAMP             234     //Ramping "strobe"


#ifdef STRIPPED  // define what you want to remove in stripped mode
#undef TEMPERATURE_MON
#undef VOLTAGE_MON
#define NO_STROBES  // don't use strobes (not stripped from modes though, likely will produce turbo)
#undef USE_BATTCHECK
#endif



/**  See modegroups.h to configure ramps and modegroups.  **/

/**********************************************************************************
**********************END OF CONFIGURATION*****************************************
***********************************************************************************/

#define GROUP_SELECT_MODE 254
#define TEMP_CAL_MODE 253


#if defined(FULL_BIKING_STROBE) && !defined(BIKING_STROBE)
#define BIKING_STROBE
#endif


#define OWN_DELAY           // Don't use stock delay functions.
#define USE_DELAY_S         // Also use _delay_s(), not just _delay_ms()

#if !defined(OFFTIM3)
  #define HIDDENMODES       // initializer, edit in modegroups.h, not here.
#endif

#include "modegroups.h"

//uint8_t modes[] = { 1,2,3,4,5,6,7,8,9, HIDDENMODES };  // make sure this is long enough...
uint8_t modes[MAX_MODES + sizeof(hiddenmodes)];  // make sure this is long enough

// Modes (gets set when the light starts up based on saved config values)
  PROGMEM const uint8_t ramp_pwm1[] = { RAMP_PWM1 };
#ifdef PWM2_LVL
  PROGMEM const uint8_t ramp_pwm2[] = { RAMP_PWM2 };
#endif
#ifdef PWM3_LVL
  PROGMEM const uint8_t ramp_pwm3[] = { RAMP_PWM3 };
#endif
#ifdef PWM4_LVL
  PROGMEM const uint8_t ramp_pwm4[] = { RAMP_PWM4 };
#endif


// Calibrate voltage and OTC in this file:
#include "fr-calibration.h"

#ifdef OTSM_powersave
#include "fr-delay.h" // must include them in this order.
#define _delay_ms _delay_sleep_ms
#define _delay_s _delay_sleep_s
#else
#include "tk-delay.h"
#endif


#ifdef RANDOM_STROBE
#include "tk-random.h"
#endif

#include "tk-voltage.h"


///////////////////////REGISTER VARIABLES///////////////////////////////////
// make it a register variable.  Saves many i/o instructions.
// but could change as code evolves.  Worth testing what works best again later.

// ************NOTE global_counter is now reserved as r2 in tk-delay.h. (well, commented out maybe)**********

register uint8_t  mode_idx asm("r6");           // current or last-used mode number
uint8_t  mode_idx;           // current or last-used mode number
register uint8_t eepos asm("r7");
//uint8_t eepos

#if defined(USE_OTSM) || defined(USE_ESWITCH)
//Reserving a register for this allows checking it in re-entrant interrupt without needing register maintenance
// as a side effect it saves several in/out operations that save more space.
register uint8_t wake_count asm  ("r8") ;
#endif
/////////////////////END REGISTERVARIABLE/////////////////////////////////

// counter for entering config mode
// (needs to be remembered while off, but only for up to half a second)
uint8_t fast_presses __attribute__ ((section (".noinit")));

//
//#if !defined (OFFTIM3) FR: Could do this to check fast_presses corruption:
//uint8_t fast_presses2 __attribute__ ((section (".noinit")));
//uint8_t fast_presses3 __attribute__ ((section (".noinit")));
//uint8_t fast_presses4 __attribute__ ((section (".noinit")));
//#endif
// but need to make a set_fast_presses/validate_fast_presses code.

// total length of current mode group's array
uint8_t mode_cnt;
// number of regular non-hidden modes in current mode group
uint8_t solid_modes;
// number of hidden modes in the current mode
// (hardcoded because both groups have the same hidden modes)
//uint8_t hidden_modes = NUM_HIDDEN;  // this is never used


#define FAST 0x03           // fast PWM
#define PHASE 0x01         // phase-correct PWM

#define n_saves 12  // number  of options to save/restore
// FR makes these all aray variables now
// can now loop to save, loop to load, and loop to toggle
// saves about 30 subtroutine calls, so a bunch of flash space.
uint8_t OPT_array[n_saves+1];
#define OPT_firstboot (EEPSIZE-1) // only one that uses individual reads.
#define muggle_mode   OPT_array[1]
#define memory        OPT_array[2]
#define enable_moon   OPT_array[3]
#define reverse_modes OPT_array[4]
//#define mode_override OPT_array[5]  // don't need to define mod_override entries here, just leave a space.
#define offtim3       OPT_array[6]
//#define mode_override OPT_array[7]
#define firstboot     OPT_array[8]  // start at 1 to match EEPSIZE math
#define lockswitch    OPT_array[9]  // don't forget to update n_saves above.
#define modegroup     OPT_array[10]
#define maxtemp       OPT_array[11]
#define _configs      OPT_array[12]

#define n_toggles 9// how many of the saves are toggle variables.
uint8_t overrides[n_toggles]; // will hold values override mode_idx if any for each toggle.

#define FIRSTBOOT 0b01010101

// define a bit field in an I/O register with sbi/cbi support, for efficient boolean manipulation:

typedef struct
{
	unsigned char bit0:1;
	unsigned char bit1:1;
	unsigned char bit2:1;
	unsigned char bit3:1;
	unsigned char bit4:1;
	unsigned char bit5:1;
	unsigned char bit6:1;
	unsigned char bit7:1;
}io_reg;
/***************LETS RESERVE GPIOR0 FOR LOCAL VARIABLE USE ********************/
// This seemed more useful at one point than it maybe actually is, but it works.
//define a global bitfield in lower 32 IO space with in/out sbi/cbi access
		#define gbit            *((volatile uint8_t*)_SFR_MEM_ADDR(GPIOR2)) // pointer to whole status byte.
		#define gbit_pwm4zero			((volatile io_reg*)_SFR_MEM_ADDR(GPIOR2))->bit0 // standby, during off-press (not on press)
		#define gbit_epress	     	((volatile io_reg*)_SFR_MEM_ADDR(GPIOR2))->bit1  // eswitch was pressed.
//      ...
        #define gbit_addr  "0x13"  // hack, but whatever, 0x13 is GPIOR2 but GPIOR2 macro is just too clever to work.
		                            //There's surely a better way, but not worth finding out.
		#define gbit_pwm4zero_bit "0"  // This gets used for inline asm in PWM ISR.

/* What about making these a bit field?  Hmm, maybe not helpful, still a read, modify, write for toggles, no sbi/sbc
*  but save/restore turns into a one or two byte thing, probably not worth the effort. */
/*
#define _configs           *((volatile uint8_t*)_SFR_MEM_ADDR(GPIOR1)) // pointer to whole status byte.
#define memory	    	    ((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit0 // now unused
#define offtim3	    	    ((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit1 // now unused
#define mode_override	    	    ((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit2 // now unused
#define enable_moon			((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit3 // standby, during off-press (not on press)
#define reverse_modes	     	((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit4
#define muggle_mode	    	    ((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit5 // now unused
#define lockswitch	    	    ((volatile io_reg*)_SFR_MEM_ADDR(GPIOR1))->bit6 // now unused
*/

#if defined(USE_OTSM) || defined(USE_ESWITCH)
// convert wake time thresholds to counts, math done by compiler, not runtime:
// plus 1 because there's always an extra un-timed pin-change wake:
#define wake_count_short (uint8_t)(1+((double)wake_time_short)*((uint16_t)1<<(6-SLEEP_TIME_EXPONENT)))
#define wake_count_med (uint8_t)(1+((double)wake_time_med)*((uint16_t)1<<(6-SLEEP_TIME_EXPONENT)))
#endif


#define Q(x) #x
#define Quote(x) Q(x)  // for placing expanded defines into inline assembly with quotes around them.
                       // the double expansion is a weird but necessary quirk, found this SO I think.

/**************WARN ON SOME BROKEN COMBINATIONS******************/

#ifdef USE_ESWITCH
   #ifndef ESWITCH_PIN
    #error You must define an ESWITCH_PIN in fr-tk-attiny.h or undefine USE_ESWITCH
   #endif
#endif

#ifdef USE_OTSM
  #ifndef OTSM_PIN
    #error You must define an OTSM_PIN in fr-tk-attiny.h or undefine USE_OTSM
  #endif
#endif

#if defined(VOLTAGE_MON)&&!(defined(READ_VOLTAGE_FROM_VCC)||defined(READ_VOLTAGE_FROM_DIVIDER))
  #error You must define READ_VOLTAGE_FROM_VCC or READ_VOLTAGE_FROM_DIVEDER when VOLTAGE_MON is defined
#endif

#if defined(VOLTAGE_MON)&&defined(READ_VOLTAGE_FROM_VCC)&&defined(READ_VOLTAGE_FROM_DIVIDER)
  #error You cannot define READ_VOLTAGE_FROM_VCC at the same time as READ_VOLTAGE_FROM_DIVIDER
#endif

#if defined(READ_VOLTAGE_FROM_DIVIDER)&&!defined(REFERENCE_DIVIDER_READS_TO_VCC)&&defined(USE_OTSM)&&(OTSM_PIN==VOLTAGE_PIN)&&defined(VOLTAGE_MON)
  #error You cannot use 1.1V reference to read voltage on the same pin as the OTSM.  OTSM power-on voltage must be >1.8V.
  #error Use REFERENCE_DIVIDER_READS_TO_VCC instead and set divider resistors to provide > 1.8V on-voltage to voltage pin.
  #error For 1S non-LDO or 1-S 5.0V LDO lights you can use READ_VOLTAGE_FROM_VCC instead
#endif

#if defined(READ_VOLTAGE_FROM_DIVIDER)&&!defined(REFERENCE_DIVIDER_READS_TO_VCC)&&defined(USE_ESWITCH)&&(ESWITCH_PIN==VOLTAGE_PIN)&&defined(VOLTAGE_MON)
  #error You cannot use 1.1V reference to read voltage on the same pin as the e-switch.  E-switch on voltage must be >1.8V.
  #error Use REFERENCE_DIVIDER_READS_TO_VCC instead and set divider resistors to provide > 1.8V on-voltage to voltage pin.
  #error For 1S non-LDO or 1-S 5.0V-LDO lights you can use READ_VOLTAGE_FROM_VCC instead.
#endif

#if !defined(USE_OTC)&&!defined(USE_OTSM)&&defined(OFFTIM3)
  #error You must define either USE_OTC or USE_OTSM when OFFTIM3 is defined.
#endif


// Not really used
//#define DETECT_ONE_S  // auto-detect 1s vs multi-s light make prev.
// requiers 5V LDO for multi-s light and will use VCC for 1S.
// incomplete, needs calibration table switching, costs memory, probably 160 bytes when done.


/****************************END OF PREPROCESSOR DEFINES**************************/

uint8_t mode_idx;
inline void initial_values() {//FR 2017
	 // configure initial values here, 255 disables the toggle. 
    #ifdef USE_FIRSTBOOT
	 firstboot = FIRSTBOOT;               // detect initial boot or factory reset
	#else
	 firstboot=255;                       // disables menu toggle
	#endif
	 modegroup = INIT_MODEGROUP;          // which mode group will be default, mode groups below start at zero, select the mode group you want and subtract one from the number to get it by defualt here
	 enable_moon = INIT_ENABLE_MOON;      // Should we add moon to the set of modes?
	 reverse_modes = INIT_REVERSE_MODES;  // flip the mode order?
	 memory = INIT_MEMORY;                // mode memory, or not
	#ifdef OFFTIM3
	 offtim3 = INIT_OFFTIM3;              // enable medium-press?
	#else
	 offtim3=255;                         //disable menu toggle, 
	#endif
	#ifdef TEMPERATURE_MON
	 maxtemp = INIT_MAXTEMP;              // temperature step-down threshold, each number is roughly equal to 4c degrees
	#endif
	 muggle_mode = INIT_MUGGLE_MODE;      // simple mode designed for mugglesotsm
	// Other state variables
	//	mode_override = 0;                // do we need to enter a special mode?
	 mode_idx=0;                          // initial mode
	 lockswitch=INIT_LOCKSWITCH;          // E-swtich enabled.
	
	 overrides[5]=GROUP_SELECT_MODE;      // defines a mode in the override range that the toggle enables.
	#ifdef TEMPERATURE_MON
	 overrides[7]=TEMP_CAL_MODE;          // defines a mode in the override range that the toggle enables.
	#else
	 OPT_array[7]=255;
	#endif
}


void save_mode() {  // save the current mode index (with wear leveling)
    uint8_t oldpos=eepos;

    eepos = (eepos+1) & ((EEPSIZE/2)-1);  // wear leveling, use next cell

    eeprom_write_byte((uint8_t *)((uint16_t)eepos),mode_idx);  // save current state
    eeprom_write_byte((uint8_t *)((uint16_t)oldpos), 0xff);     // erase old state
}


#define WRITE 1
#define READ 0


void getsetstate(uint8_t rw){// double up the function to save space.
   uint8_t i;
   i=n_saves;
   do{
       if (rw==WRITE){// conditional inside loop for space.
	       eeprom_write_byte((uint8_t *)(EEPSIZE-i),OPT_array[i]);
	   } else {
	       OPT_array[i]=eeprom_read_byte((uint8_t *)(EEPSIZE-i));
       }
	   i--;
    } while (i); // don't include 0
}


void save_state() {  // central method for writing complete state
	save_mode();
	getsetstate(WRITE);
}


#ifndef USE_FIRSTBOOT
inline void reset_state() {
    mode_idx = 0;
    modegroup = 5;
    save_state();
}
#endif

inline void restore_state() {
    uint8_t eep;

#ifdef USE_FIRSTBOOT
     //check if this is the first time we have powered on
    eep = eeprom_read_byte((uint8_t *)OPT_firstboot);
    if (eep != FIRSTBOOT) {// confusing: if eep = FIRSTBOOT actually means the first boot already happened
		                   // so eep != FIRSTBOOT means this IS the first boot.
        // not much to do; the defaults should already be set
        // while defining the variables above
        save_state(); // this will save FIRSTBOOT to the OPT_firstboot byte.
        return;
    }
#endif
    eepos=0; 
    do { // Read until we get a result. Tightened this up slightly -FR
	    mode_idx = eeprom_read_byte((const uint8_t *)((uint16_t)eepos));
    } while((mode_idx == 0xff) && eepos++<(EEPSIZE/2) ) ; // the && left to right eval prevents the last ++.
#ifndef USE_FIRSTBOOT

    // if no mode_idx was found, assume this is the first boot
    if (mode_idx==0xff) {
//    if (first) {
        reset_state();
        return;
    }
#endif
    // load other config values
    getsetstate(READ);

#ifndef USE_FIRSTBOOT
    if (modegroup >= NUM_MODEGROUPS) reset_state();
#endif
}

inline void next_mode() {
    mode_idx += 1;
    if (mode_idx >= solid_modes) {
        // Wrap around, skipping the hidden modes
        // (note: this also applies when going "forward" from any hidden mode)
        // FIXME? Allow this to cycle through hidden modes?
        mode_idx = 0;
    }
}

#ifdef OFFTIM3
inline void prev_mode() {
    // simple mode has no reverse
    if (muggle_mode) { return next_mode(); }

    if (mode_idx == solid_modes) { 
        // If we hit the end of the hidden modes, go back to first mode
        mode_idx = 0;
    } else if (mode_idx > 0) {
        // Regular mode: is between 1 and TOTAL_MODES
        mode_idx -= 1;
    } else {
        // Otherwise, wrap around (this allows entering hidden modes)
        mode_idx = mode_cnt - 1;
    }
}
#endif



inline void count_modes() {
    /*
     * Determine how many solid and hidden modes we have.
     *
     * (this matters because we have more than one set of modes to choose
     *  from, so we need to count at runtime)
     */
	/*  
	 * FR: What this routine really does is combine the moon mode, the defined group modes, 
	 * and the hidden modes into one array according to configuration options
	 * such as mode reversal, enable mood mode etc.
	 * The list should like this:
	 * modes ={ moon, solid1, solid2,...,last_solid,hidden1,hidden2,...,last_hidden}
	 * unless reverse_modes is enabled then it looks like this:
	 * modes={last_solid,last_solid-1,..solid1,moon,hidden1,hidden2,...last_hidden}
	 * or if moon is disabled it is just left out of either ordering.
	 * If reverse clicks are disabled the hidden modes are left out.
	 * mode_cnt will hold the final count of modes  -FR.
	 */	 
    // copy config to local vars to avoid accidentally overwriting them in muggle mode
    // (also, it seems to reduce overall program size)
    uint8_t my_modegroup = modegroup;
    uint8_t my_enable_moon = enable_moon;
    uint8_t my_reverse_modes = reverse_modes;

    // override config if we're in simple mode
    if (muggle_mode) {
        my_modegroup = NUM_MODEGROUPS;
        my_enable_moon = 0;
        my_reverse_modes = 0;
    }
    //my_modegroup=11;
    //my_reverse_modes=0; // FOR TESTING
	//my_enable_moon=1;
	uint8_t i=0;
//    const uint8_t *src = modegroups + (my_modegroup<<3);
    const uint8_t *src=modegroups;
	// FR adds 0-terminated variable length mode groups
	// scan for enough 0's, leave the pointer on the first mode of the group.
    i=my_modegroup;
	while(i) {
			while(pgm_read_byte(src++)){}
				i--;
	}
	
    // Figure out how many modes are in this group
	solid_modes=0;
	//FR:  do we really need the safety check?  It's a check on the developer, not user input. 
//    while((solid_modes<=MAX_MODES) && pgm_read_byte(src+solid_modes)){
	  while(pgm_read_byte(src+solid_modes)){// this version saves 6 bytes.
//TODO:  if modes[idx]  was turned into modes(idx);	there woulnd't be a limit anyway.
	    solid_modes++; 
    }
    i=0;
    // Now handle reverse order in same step. -FR
	mode_cnt=solid_modes;

	do{
     //note my_enable_moon and my_reverse_modes are either 1 or 0, unlike solid_modes.
    modes[(my_reverse_modes?solid_modes-mode_cnt:mode_cnt-1+my_enable_moon)] = pgm_read_byte(src+(mode_cnt-1));
     mode_cnt--;
    } while (mode_cnt);

	solid_modes+=my_enable_moon;

    if (my_enable_moon) {// moon placement already handled, but have to fill in the value.
	    modes[(my_reverse_modes?solid_modes-1:0)] = 1;
    }

// add hidden modes
#ifdef OFFTIM3
    mode_cnt = solid_modes; // statement is redundant, but saves bytes anyway.

    for( i=sizeof(hiddenmodes);i>0; i-- ) // not clear if reversing this saved memory or just using i did.
    {
	    modes[mode_cnt++] = pgm_read_byte((uint8_t *)(hiddenmodes+sizeof(hiddenmodes)-i));
    }
#else
mode_cnt = solid_modes; 
#endif
// remove last hidden mode if needed to avoid turbo to turbo.
	mode_cnt-=!(modes[0]^modes[mode_cnt-1]);// cheapest syntax I could find, but the check still costs 20 bytes. -FR
}

// This is getting too silly
// and tk already removed all usages of it anyway, except from set_level, so just do it there.
//#ifdef PWM4_LVL
  //inline void set_output(uint8_t pwm1, uint8_t pwm2, uint8_t pwm3, uint8_t pwm4) {
//#elif defined(PWM3_LVL)
  //inline void set_output(uint8_t pwm1, uint8_t pwm2, uint8_t pwm3) {
//#elif defined(PWM2_LVL)
  //inline void set_output(uint8_t pwm1, uint8_t pwm2) {
//#else
  //inline void set_output(uint8_t pwm1,) {
//#endif
//#endif
  //inline void set_output() {
    ///* This is no longer needed since we always use PHASE mode.
    //// Need PHASE to properly turn off the light
    //if ((pwm1==0) && (pwm2==0)) {
        //TCCR0A = PHASE;
    //}
    //*/
    //FET_PWM_LVL = pwm1;
    //PWM_LVL = pwm2;
//#ifdef ALT_PWM_LVL
      //ALT_PWM_LVL = pwm3;
//#endif
////return;
//}

void set_level(uint8_t level) {
    if (level == 0) {
#ifdef PWM4_LVL
        gbit_pwm4zero=1; // we can't disable the interrupt because we use it for other timing
		                 // so have to signal the interrupt instead.
        PWM4_LVL=0;
#endif
#if defined(PWM3_LVL)
        PWM3_LVL=0;
#endif
#if defined(PWM2_LVL)
        PWM2_LVL=0;
#endif
#ifdef PWM_LVL
        PWM_LVL=0;
#endif
    } else {
        level -= 1;
#ifdef PWM4_LVL
       PWM4_LVL=pgm_read_byte(ramp_pwm4   + level);
       gbit_pwm4zero=0;
#endif
#ifdef PWM3_LVL
       PWM3_LVL=pgm_read_byte(ramp_pwm3   + level);
#endif
#ifdef PWM2_LVL
       PWM2_LVL=pgm_read_byte(ramp_pwm2   + level);
#endif
#ifdef PWM_LVL
       PWM_LVL=pgm_read_byte(ramp_pwm1   + level);
#endif
    }
}


inline void set_mode(uint8_t mode) {
#ifdef SOFT_START
    static uint8_t actual_level = 0;
    uint8_t target_level = mode;
    int8_t shift_amount;
    int8_t diff;
    do {
        diff = target_level - actual_level;
        shift_amount = (diff >> 2) | (diff!=0);
        actual_level += shift_amount;
        set_level(actual_level);
        _delay_ms(RAMP_SIZE/20);  // slow ramp
        //_delay_ms(RAMP_SIZE/4);  // fast ramp
    } while (target_level != actual_level);
#else
#define set_mode set_level
    //set_level(mode);
#endif  // SOFT_START
}

void blink(uint8_t val, uint16_t speed)
{
    for (; val>0; val--)
    {
        set_level(BLINK_BRIGHTNESS);
        _delay_ms(speed);
        set_level(0);
        _delay_ms(speed);
        _delay_ms(speed);
    }
}

// utility routine to blink out 8bit value as a 3 digit decimal.
// The compiler will omit it if it's not used.  Good for debugging.  -FR.
// uses much memory though.
inline void blink_value(uint8_t value){
	blink (10,20);
	_delay_ms(500);
	blink((value/100)%10,200); // blink hundreds
	_delay_s();
	blink((value/10)%10,200); // blink tens
	_delay_s();
	blink(value % 10,200); // blink ones
	_delay_s();
	blink(10,20);
	_delay_ms(500);
}


void strobe(uint8_t ontime, uint8_t offtime) {
// TK added this loop.  The point seems to be to slow down one strobe cycle
// so fast_presses timeout is not reached too quickly.
// FR has a partial implementation of a global clock in tk-delay
// that would solve this more generally, but for now this is slightly smaller and does the job.
    uint8_t i;
    for(i=0;i<8;i++) {
	    set_level(RAMP_SIZE);
	    _delay_ms(ontime);
	    set_level(0);
	    _delay_ms(offtime);
    }
}


#ifdef SOS
inline void SOS_mode() {
#define SOS_SPEED 200
    blink(3, SOS_SPEED);
    _delay_ms(SOS_SPEED*5);
    blink(3, SOS_SPEED*5/2);
    //_delay_ms(SOS_SPEED);
    blink(3, SOS_SPEED);
    _delay_s(); _delay_s();
}
#endif

#ifdef NEVER_USED
void toggle(uint8_t *var, uint8_t num) {
    // Used for config mode
    // Changes the value of a config option, waits for the user to "save"
    // by turning the light off, then changes the value back in case they
    // didn't save.  Can be used repeatedly on different options, allowing
    // the user to change and save only one at a time.
    blink(num, BLINK_SPEED/8);  // indicate which option number this is
    *var ^= 1;
    save_state();
    // "buzz" for a while to indicate the active toggle window
    blink(32, 500/32);
    /*
    for(uint8_t i=0; i<32; i++) {
        set_level(BLINK_BRIGHTNESS * 3 / 4);
        _delay_ms(20);
        set_level(0);
        _delay_ms(20);
    }
    */
    // if the user didn't click, reset the value and return
    *var ^= 1;
    save_state();
    _delay_s();
}
#endif

void toggles() {
    // New FR version loops over all toggles, avoids variable passing, saves ~50 bytes.
    // mode_overrides are handled by an array entry for each toggle that sets mod_ix
	// to either 0 or an override mode.  On startup mode_idx is checked to be in the 
	// override range or not.
	//
    // Used for config mode
    // Changes the value of a config option, waits for the user to "save"
    // by turning the light off, then changes the value back in case they
    // didn't save.  Can be used repeatedly on different options, allowing
    // the user to change and save only one at a time.
    uint8_t i=0;
	do {	
		 i++;
		 if(OPT_array[i]!=255){  
	       blink(i, BLINK_SPEED/8);  // indicate which option number this is
	       mode_idx=overrides[i]; 
           OPT_array[i] ^= 1;
           save_state();
    // "buzz" for a while to indicate the active toggle window
           blink(32, 500/32);
           OPT_array[i] ^= 1;
           save_state();
           _delay_s();
		   if (muggle_mode) {break;} // go through once on muggle mode
			                         // this check adds 10 bytes.
		 }
    }while((i<n_toggles));
}



#ifdef TEMPERATURE_MON
uint8_t get_temperature() {
    // average a few values; temperature is noisy
    uint16_t temp = 0;
    uint8_t i;
    ADC_on_temperature(); // do it before calling, inconsistent with get_voltage and
                           
    for(i=0; i<16; i++) {
        _delay_ms(5);  // 
        temp += get_voltage();
    }
    temp >>= 4;
    return temp;
}
#endif  // TEMPERATURE_MON

#if defined(OFFTIM3)&&defined(USE_OTC)
inline uint8_t read_otc() {
    // Read and return the off-time cap value
    // Start up ADC for capacitor pin
    // disable digital input on ADC pin to reduce power consumption
    //DIDR0 |= (1 << CAP_DIDR); // will just do it for all pins on init.
    // 1.1v reference, left-adjust, ADC3/PB3
    ADMUX  = (1 << V_REF) | (1 << ADLAR) | CAP_PIN;
    // enable, start, prescale
    ADCSRA = (1 << ADEN ) | (1 << ADSC ) | ADC_PRSCL;
    // Wait for completion
    while (ADCSRA & (1 << ADSC));
    // Start again as datasheet says first result is unreliable
    ADCSRA |= (1 << ADSC);
    // Wait for completion
    while (ADCSRA & (1 << ADSC));

    // ADCH should have the value we wanted
    return ADCH;
}
#endif

inline void init_mcu(){ // complete revamp by FR to allow use of any combination of 4 outputs, OTSM, and more
	// 2017 by Flintrock (C)
	//
	// setup all our initial mcu configuration, mostly including interrupts.
    //DDRB, 1 bit per pin, 1=output default =0.
	//PORTB 1 is ouput-high or input-pullup.  0 is output-low or input floating/tri-stated.  Default=0.
	//DIDR0 Digital input disable, 1 disable, 0 enable. disable for analog pins to save power. Default 0 (enabled) 

    // start by disabling all digital input.
    // PWM pins will get set as output anyway.
    DIDR0=0b00111111 ; // tests say makes no difference during sleep,
	                   // but datasheet claims it helps drain in middle voltages, so during power-off detection maybe.
	
	// Charge up the capacitor by setting CAP_PIN to output
#ifdef CAP_PIN
    // if using OTSM but no OTC cap, probably better to stay input high on the unused pin.
	// If USE_ESWITCH and eswitch pin is defined same as cap pin, the eswitch define overrides, 
	//  so again, leave it as input (but not high).
    #if ( (defined(USE_OTSM) && !defined(OTSM_USES_OTC) ) || (defined(USE_ESWITCH) && ESWITCH_PIN == CAP_PIN ) )
       // just leave cap_pin as an input pin (raised high by default lower down)
    #else
	// Charge up the capacitor by setting CAP_PIN to output
	  DDRB  |= (1 << CAP_PIN);    // Output
	  PORTB |= (1 << CAP_PIN);    // High
    #endif
#endif
	
	// Set PWM pins to output
    #ifdef PWM_PIN
  	   DDRB |= (1 << PWM_PIN) ;    
	#endif
	#ifdef PWM2_PIN
	   DDRB |= (1 << PWM2_PIN);
	#endif 
    #ifdef PWM3_PIN
  	   DDRB |= (1 << PWM3_PIN); 
    #endif
    #ifdef PWM4_PIN
       DDRB |= (1 << PWM4_PIN); 
    #endif
	  
	
	// Set Timer 0 to do PWM1 and PWM2 or for OTSM powersave timing
    #if defined(PWM_PIN) || defined (PWM2_PIN) || defined OTSM_powersave // should alias the last one to something better 	
     
 	   TCCR0A = PHASE; // phase means counter counts up and down and 
	               // compare register triggers on and off symmetrically at same value.
				   // This doubles the time so half the max frequency.
	// Set prescaler timing
	// This and the phase above now directly impact the delay function in tk-delay.h..
	//  make changes there to compensate if this is changed
  	     TCCR0B = 0x01; // pre-scaler for timer (1 => 1, 2 => 8, 3 => 64...)  1 => 16khz in phase mode or 32 in fast mode.

       #ifdef  PWM_PIN
	     TCCR0A |= (1<<COM0B1);  // enable PWM on PB1
	   #endif
	   #ifdef PWM2_PIN
	     TCCR0A |= (1<<COM0A1);  // enable PWM on PB0
	   #endif
	#endif       

   //Set Timer1 to do PWM3 and PWM4
	#if defined(PWM3_PIN) || defined (PWM4_PIN)
	  OCR1C = 255;  
	// FR: this timer does not appear to have any phase mode.
	// The frequency appears to be just system clock/255
	#endif
	#ifdef PWM3_PIN
	  TCCR1 = _BV (CS10);  // prescale of 1 (32 khz PWM)
	  GTCCR = (1<<COM1B1) | (1<<PWM1B) ; // enable pwm on OCR1B but not OCR1B compliment (PB4 only). 
	                                    // FR: this timer does not appear to have any phase mode.
										// The frequency appears to be just system clock/255
    #endif										
	#ifdef PWM4_PIN
	  TCCR1 = _BV (CS11);  // override, use prescale of 2 (16khz PWM) for PWM4 because it will have high demand from interrupts.
	                       // This will still be just as fast as timer 0 in phase mode.
  	  TCCR1 |=  _BV (PWM1A) ;           // for PB3 enable pwm for OCR1A but leave it disconnected.  
		                                // Its pin is already in use.
							            // Must handle it with an interrupt.
      TIMSK |= (1<<OCIE1A) | (1<<TOIE1);        // enable interrupt for compare match and overflow on channel 4 to control PB3 output
	  // Note: this will then override timer0 as the delay clock interrupt for tk-delay.h.
	#endif
	
		
	
	PORTB=~DDRB; // Anything that is not an output gets a pull up resistor.
	// again, no change measured during sleep, but maybe helps with shutdown?
	// exceptions handled below.
#ifdef USE_ESWITCH
	 // leave pullup set on E-switch pin
	  PCMSK |= 1<< ESWITCH_PIN;  // catch pin change on eswitch pin
#endif
#if defined(USE_OTSM) 
//  no pull-up on OTSM-PIN. 
      PORTB=PORTB&~(1<<OTSM_PIN); 
	  PCMSK |= 1<< OTSM_PIN;  // catch pin change on OTSM PIn
#endif
#if defined(USE_ESWITCH) || defined(USE_OTSM)
	GIMSK |= 1<<PCIE;	// enable PIN change interrupt

	sleep_enable();     // just leave it enabled.  It's fine and will help elsewhere.
	GIFR |= PCIF;  //    Clear pin change interrupt flag
	sei();				//Enable Global Interrupt
#endif

#if (OTSM_PIN != VOLTAGE_PIN)  && !defined(READ_VOLTAGE_FROM_VCC) // also true if OTSM_PIN=VOLTAGE_PIN,  but redundant.
    PORTB=PORTB&~(1<<VOLTAGE_PIN); //pull up all non-outputs except voltage pin
	// note if the eswitch is on the voltage pin, and there are dividers, this will break the e-switch.
	// must comment out the voltage pin in attiny.h then.
#endif


    //timer overflow interrupt is presently setup in tk-delay.h for OTS_powersave

}


#if defined(USE_OTSM) || defined(USE_ESWITCH)
// decide which mode to switch to based on user input
inline void change_mode() {  // just use global wake_count variable
#else
inline void change_mode(uint8_t cap_val) {
#endif
//	if (! mode_override) {
  if (mode_idx<MINIMUM_OVERRIDE_MODE) {
   #if defined(USE_OTSM) || defined(USE_ESWITCH)
	  // wake_time is in units of 64ms (2^6 ms) per tick.
	  // The compiler will do the floating point conversion here
	    if (wake_count < wake_count_short   )  { // plus 64 (2^6) to round up.
   #elif defined(USE_OTC)		
		if (cap_val > CAP_SHORT) {
   #else
		if (fast_presses < 0x20) { // fallback, but this trick needs improvements.
   #endif
				// Indicates they did a short press, go to the next mode
				// We don't care what the fast_presses value is as long as it's over 15
				fast_presses = (fast_presses+1) & 0x1f;
				next_mode(); // Will handle wrap arounds

		  #ifdef OFFTIM3
            #if defined(USE_OTSM) || defined(USE_ESWITCH)
				} else if (wake_count < wake_count_med   )  {
            #else
				} else if (cap_val > CAP_MED) {
            #endif

				// User did a medium press, go back one mode
				fast_presses = 0;
				if (offtim3) {
					prev_mode();  // Will handle "negative" modes and wrap-arounds
					} else {
					next_mode();  // disabled-med-press acts like short-press
					// (except that fast_presses isn't reliable then)
				}
		  #endif
				} else {
				// Long press, keep the same mode
				// ... or reset to the first mode
				fast_presses = 0;
				if (muggle_mode  || (! memory)) {
					// Reset to the first mode
					mode_idx = 0;
				}
			}
		}

}

#ifdef VOLTAGE_CAL
inline void voltage_cal(){
	ADC_on();
	while(1){
  	  _delay_ms(300);
	  blink_value(get_voltage());
	}
}
#endif


/*********************ISRs***************************************************/
/******One more is in FR-delay.h*******************************************/

#ifdef PWM4_LVL
// 2017 by Flintrock.
// turn on PWM4_PIN at bottom of clock ramp
// this will also replace the delay-clock interrupt used in fr-delay.h
// Since SBI,CBI, and SBIC touch no registers, so no register protection is needed:
// PWM this way without a phase-mode counter would not shut off the light completely
// Added a control flag for 0 level to disable the light. We need the interrupts running for the delay clock.

ISR(TIMER1_OVF_vect,ISR_NAKED){ // hits when counter 1 reaches max
	  __asm__ __volatile__ (    //0x18 is PORTB
//	  "sbic  " gbit_addr " , " gbit_pwm4zero_bit " \n\t" 
//	  "rjmp .+2 " "\n\t"
	  "cbi 0x18 , " Quote(PWM4_PIN) " \n\t"
	  "reti" "\n\t"
	  :::);
}

// turn off PWM4_PIN at COMPA register match
// this one is only defined here though:
ISR(TIMER1_COMPA_vect, ISR_NAKED){ //hits when counter 1 reaches compare value
//	if(!gbit_pwm4zero) {
	__asm__ __volatile__ (    //0x18 is PORTB
	  "sbic  " gbit_addr " , " gbit_pwm4zero_bit " \n\t" // if pwm4zero isn't set...
      "rjmp .+2 " "\n\t"
  	  "sbi 0x18 , " Quote(PWM4_PIN) " \n\t"  //set bit in PORTB (0x18) to make PW4_PIN high.
	  "reti" "\n\t"
	  :::);
//	}
}
#endif

#if defined(USE_OTSM) || defined(USE_ESWITCH)
// Off-time sleep mode by -FR, thanks to flashy Mike and Mike C for early feasibility test with similar methods.

//need to declare main here to jump to it from ISR.
int main();

// OS_task is a bit safer than naked,still allocates locals on stack if needed, although presently avoided.
void __attribute__((OS_task, used, externally_visible)) PCINT0_vect (void) { 
//ISR(PCINT0_vect, ISR_NAKED){// Dual watchdog pin change power-on wake by FR

	// 2017 by Flintrock (C)
    //
//PIN change interrupt.  This serves as the power down and power up interrupt, and starts the watchdog.
//
	/* IS_NAKED (or OS_task) IS ***DANGEROUS** remove it if you edit this function significantly
	 *  and aren't sure. You can remove it, and change the asm reti call below to a return.
	 * and remove the CLR R1. However, it saves about 40 bytes, which matters.
	 * The idea here, the compiler normally stores all registers used by the interrupt and
	 * restores them at the end of the interrupt.  This costs instructions (space).
     * We won't return to the caller anyway, we're going to reset (so this is similar to a 
	 * tail call optimization).
	 * One tricky bit then is that we  get a re-entrant call here because 
	 * sleep and wake use the same interrupt, so we'll interrupt the sleep command below with a re-entry
	 * into this function on wake. The re-entry will pass the "if wake_count" conditional and immediately 
	 * return to the original instance. 
	 * We avoid touching much there by using a global register variable for wake_count.
	 * SREG is probably changed but since it only breaks out of the sleep there are
	 * no pending checks of result flags occurring. 
	 *
	 * Another issue is allocation of locals. Naked has no stack frame,
	 * but seems to properly force variables to register (better anyway) until it can't, 
	 * and then fails to compile (gcc 4.9.2) OS_task however will create a stack frame if needed and is 
	 * more documented/gauranteed/future-proof probably.
	 */

	__asm__ __volatile__ ("CLR R1"); // We do need a clean zero-register (yes, that's R1).
	#ifdef USE_ESWITCH // for eswtich we'll modify the sleep loop after a few seconds.
	    register uint8_t sleep_time_exponent=SLEEP_TIME_EXPONENT; 
	#else
	    #define sleep_time_exponent SLEEP_TIME_EXPONENT
	#endif
	if( wake_count ){// if we've already slept, so we're waking up. 
		             // This is a register variable; no need to clean space for it.
//		return; // we are waking up, do nothing and continue from where we slept (below).
		__asm__ __volatile__ ("reti");
	} // otherwise, we're  going to sleep
// we only pass this section once, can do initialization here:	
//    e_status=0b00000001;   	 //startoff in standby with pin low (button pressed)
#ifdef USE_ESWITCH
     register uint8_t e_standdown=1; // pressing off, where we start
     register uint8_t e_standby=0;  // pressing on
 	 register uint8_t pins_state=0; // state of all switch pins.
     register uint8_t old_pins_state=0; // detect a change (a sane one at least).
#endif
	cli();
	ADCSRA = 0; //disable the ADC.  We're not coming back, no need to save it.
	set_level(0); // turn off the light. Is this necessary? Helpful? Saves power on PWM pins?
	              // If in eswitch, could skip it, but depends which button was clicked.
	while(1) { // keep going to sleep until power is on
    	set_sleep_mode(SLEEP_MODE_PWR_DOWN);
		// Wake on either a power-up (pin change) or a watchdog timeout
		// At every watchdog wake, increment the wake counter and go back to sleep.
// 		WDTCR = (1<<WDCE) |(1<<WDE);
//		WDTCR = (0<<WDE);  // might already be in use, so we have to clear it.
		WDTCR = (1<<WDIE) |(!!(sleep_time_exponent&8)<<WDP3)|(sleep_time_exponent&7<<WDP0); // set the watchdog timer: 2^n *16 ms datasheet pg 46 for table.
		sleep_bod_disable();
		sei(); // yep, we're going to nest interrupts, even re-entrenty into this one!
		sleep_cpu();
	         	// return from wake ISR's here, both watchdog and pin change.
		cli();
		// Now just figure out what pins are pressed and if we're coming or going, not as simple as it sounds.
	 #ifdef USE_ESWITCH
	 /*  Three phases, 
	   * 1:standown: Still armed for short click restart on button release, counting time.
	   * 2:full off (neither standby nor standown). Long press time-out ocurerd, 
	   *           -button release does nothing on eswitch.
	   *           -button press arms standby mode.
	   * 3:standby:  Button is pressed for turn on, restart on release. No timing presently.
	   * This might be more optimizable. 
	   */
	  
	    wake_count+= e_standdown; // increment click time if still in standdown mode.
	      old_pins_state = pins_state;
       #if defined(USE_OTSM) && (OTSM_PIN != ESWITCH_PIN) 
	      register uint8_t e_pin_state = (lockswitch||(PINB&(1<<ESWITCH_PIN))); 
	      pins_state = e_pin_state&&(PINB&(1<<OTSM_PIN)); // if either switch is "pressed" (0) this is 0.
	   #else
	      pins_state=(PINB&(1<<ESWITCH_PIN));
		  #if !defined(USE_OTSM)  // just an alias in this case to merge code paths:
  		    register uint8_t e_pin_state=pins_state; 
		  #endif
	   #endif
	    if ( e_standdown){			
		   if ( pins_state ) {// if button released and still in standown, restart 
		       break; // goto reset 
          #if defined(USE_OTSM) && (OTSM_PIN == ESWITCH_PIN)
			} else if ( (wake_count>wake_count_med)) { // it's a long press
 		 	    e_standdown=0; // go to full off, this will stop the wake counter.
			   _delay_loop_2(F_CPU/400); // if this was a clicky switch.. this will bring it down, 
			                             // don't want to come on in off mode.
			  sleep_time_exponent=9; // 16ms*2^9=8s, maximum watchod timeout.
			}
          #else 
			} else if (wake_count>wake_count_med && (!e_pin_state)) { // it's a long e-switch press
 		 	  e_standdown=0; // only go full off on e-switch press, count on cap to prevent counter overflow
				            // for clickly.
			  sleep_time_exponent=9; // 16ms*2^9=8s, maximum watchod timeout.
			}
		  #endif
		continue; // still in standby, go back to sleep.
	    }
        if (e_standby && pins_state  ) { // if button release while in standby, restart
	        break; // goto reset         //could merge this with e_standdown & pins_state above, might be smaller, might not.
        }
        if (!e_standdown && !pins_state && (old_pins_state)  ) { // if button pressed while off, restart
			   e_standby=1; // going to standy, will light when button is released		
		} 
		// continue.. sleep more.
	 #else  // for simple OTSM, if power comes back, we're on:
		wake_count++;		
		if(PINB&(1<<OTSM_PIN)) {
			 break; // goto reset
	    }  
	 #endif
     }	 
	 /*** Do a soft "reset" of sorts. **/
     //
     SPL = (RAMEND&0xff); // reset stack pointer, so if we keep going around we don't get a stack overflow. 
#if (ATTINY == 45 ) || ( ATTINY == 85 )
	 SPH = RAMEND>>8; 
#endif
	 SREG=0;  // clear status register
     main(); // start over, bypassing initialization.
		
} 

ISR(WDT_vect, ISR_NAKED) // watchdog interrupt handler -FR
{ // gcc avr stacks empty interrupts with a bunch of un-needed initialization.
  // do this in assembly and we get one instruction instead about 15.
        __asm__ volatile("reti");
}

#endif // OTSM




// initializer added by Flintrock.  This goes into a "secret" code segment that gets run before main. 
// It cannot be called. It's cheaper than breaking main into another sub.
// It initializes "cold boot" values. A "reset" instead jumps to main after a switch press
// and must then bypass the wake_counter initializer.
// be careful in here, local variables are not technically allowed (no stack exists), but subroutine calls are.
void __attribute__ ((naked)) __attribute__ ((section (".init8"))) \
main_init (void) { 
	 initial_values(); // load defaults for eeprom config variables.
   // initialize register variables, since gcc stubornly won't do it for us.
 	 eepos=0;
#if defined(USE_OTSM) || defined(USE_ESWITCH)
     wake_count=255; // initial value on cold reboot, represents large cick time.
#endif
}

int main(){ // Jump here from a wake interrupt, skipping initialization of wake_count.

// read OTC if needed, and translate it to an OTSM wake_count value if using both:

#if defined(OFFTIM3)&&defined(USE_OTC) 
   #if !(defined(USE_OTSM)||defined(USE_ESWITCH))
    // check the OTC immediately before it has a chance to charge or discharge
      uint8_t cap_val = read_otc();
   #elif defined(USE_ESWITCH)&!defined(USE_OTSM) 
     if(wake_count==255) { // so we had a timed out clicky press, must use OTC
		                   // translate OTC readings to wake_count
						   // could simplify and do this for normal OTC too (adds a few lines, but for a simple build anyway).
		 uint8_t cap_val = read_otc();
		 if (cap_val>CAP_MED) {
			 wake_count=wake_count_med; // long press (>= is long)
		 }else if (cap_val>CAP_SHORT){
			 wake_count=wake_count_short; // med press
		 }else{
			 wake_count=0;               // short press		     
	     }
	 }
   #elif defined(USE_OTSM)
     #error You cannot define USE_OTSM and USE_OTC at the same time.
   #endif
#endif

    init_mcu(); // setup pins, prescalers etc, configure/enable interrupts, etc.

#ifdef VOLTAGE_CAL
	voltage_cal();
#endif


#ifdef OTSM_debug // for testing your OTSM hardware 
                 // blink out the wake count, then continue into mode as normal.                 
    uint8_t temp=wake_count;
	wake_count=0; // this just resets wake_count while blinking so we can click off early and it still works.
	              // use 0 for debugging but will start at 255 in real use.
    blink_value(temp);
//    wake_time=0;  // use this to make every click act like a fast one for debugging.
#endif
    restore_state(); // loads user configurations

    count_modes(); // build the working mode group using configured choices.
 

#if defined(USE_OTSM) || defined(USE_ESWITCH)
	change_mode(); // advance, reverse, stay put?
	wake_count=0; // reset the wake counter.
#else
	change_mode(cap_val); // advance, reverse, stay put?
#endif

    // TODO: Enable this?  (might prevent some corner cases, but requires extra room)
    // memory decayed, reset it
    // (should happen on med/long press instead
    //  because mem decay is *much* slower when the OTC is charged
    //  so let's not wait until it decays to reset it)
    //if (fast_presses > 0x20) { fast_presses = 0; }
    //FR comments, if this decay is still used by anyone, it's probably best to check 4 copies 
	//  for equality (if a=b&c&d;) to detect decay
	// assuming final decay states are random, there's a 2^24 chance of a wrong answer. 
    // Tests with only one byte show that byte always decays to the same (but random seeming) value.
    // But it's probably a fixed but random value and different from other bytes?

	save_mode();
	_delay_ms(1);
#ifdef VOLTAGE_MON
    ADC_on(); 
#endif


    uint8_t output;
    uint8_t actual_level;
#ifdef TEMPERATURE_MON
    uint8_t overheat_count = 0;
#endif
    uint8_t i = 0;
#ifdef VOLTAGE_MON
    uint8_t lowbatt_cnt = 0;
#endif
    output = modes[mode_idx];
    actual_level = output;
    // handle mode overrides, like mode group selection and temperature calibration
//    if (mode_override) {
    if (mode_idx>=MINIMUM_OVERRIDE_MODE) {
        // do nothing; mode is already set
        //mode_idx = mode_override;
        fast_presses = 0;
        output = mode_idx;
    }
    while(1) {
         if (fast_presses > 0x0f) {  // Config mode
            _delay_s();       // wait for user to stop fast-pressing button
            fast_presses = 0; // exit this mode after one use
//            mode_idx = 0;
			toggles(); // this is the main menu
// The old way:	(commented out at bottom of file)	

            mode_idx = 0;
            output = modes[mode_idx];
            actual_level = output;
        }
#if !defined(NO_STROBES)

// Tried saving some space on these strobes, but it's not easy. gcc does ok though.
// Using a switch does create different code (more jump-table-ish, kind of), but it's not shorter.
#ifdef STROBE_10HZ
        else if (output == STROBE_10HZ) {
            // 10Hz tactical strobe
            strobe(33,67);
        }
#endif // ifdef STROBE_10HZ
#ifdef STROBE_16HZ
        else if (output == STROBE_16HZ) {
            // 16.6Hz tactical strobe
            strobe(20,40);
        }
#endif // ifdef STROBE_16HZ
#ifdef STROBE_8HZ
        else if (output == STROBE_8HZ) {
            // 8.3Hz Tactical strobe
            strobe(40,80);
        }
#endif // ifdef STROBE_8HZ
#ifdef STROBE_OLD_MOVIE
        else if (output == STROBE_OLD_MOVIE) {
            // Old movie effect strobe, like you some stop motion?
            strobe(1,41);
        }
#endif // ifdef STROBE_OLD_MOVIE
#ifdef STROBE_CREEPY
        else if (output == STROBE_CREEPY) {
            // Creepy effect strobe stop motion effect that is quite cool or quite creepy, dpends how many friends you have with you I suppose.
            strobe(1,82);
        }
#endif // ifdef STROBE_CREEPY

#ifdef POLICE_STROBE
        else if (output == POLICE_STROBE) {

            // police-like strobe
            //for(i=0;i<8;i++) {
                 strobe(20,40);
            //}
            //for(i=0;i<8;i++) {
                 strobe(40,80);
            //}
        }
#endif // ifdef POLICE_STROBE
#ifdef RANDOM_STROBE
        else if (output == RANDOM_STROBE) {
            // pseudo-random strobe
            uint8_t ms = 34 + (pgm_rand() & 0x3f);
            strobe(ms, ms);
            //strobe(ms, ms);
        }
#endif // ifdef RANDOM_STROBE
#ifdef BIKING_STROBE
        else if (output == BIKING_STROBE) {
            // 2-level stutter beacon for biking and such
#ifdef FULL_BIKING_STROBE
            // normal version
            for(i=0;i<4;i++) {
                set_level(TURBO);
                //set_output(255,0,0);
                _delay_ms(5);
                set_level(ONE7135);
                //set_output(0,0,255);
                _delay_ms(65);
            }
            _delay_ms(720);
#else
            // small/minimal version
            set_level(TURBO);
            //set_output(255,0,0);
            _delay_ms(10);
            set_level(ONE7135);
            //set_output(0,0,255);
            _delay_s();
#endif
        }
#endif  // ifdef BIKING_STROBE
#ifdef SOS
        else if (output == SOS) { SOS_mode(); }
#endif // ifdef SOS
#ifdef RAMP
        else if (output == RAMP) {
            int8_t r;
            // simple ramping test
            for(r=1; r<=RAMP_SIZE; r++) {
                set_level(r);
                _delay_ms(40);
            }
            for(r=RAMP_SIZE; r>0; r--) {
                set_level(r);
                _delay_ms(40);
            }
        }
#endif  // ifdef RAMP
#endif // if !defined(NO_STROBES)S
#ifdef USE_BATTCHECK
        else if (output == BATTCHECK) {
#ifdef BATTCHECK_VpT
			//blink_value(get_voltage()); // for debugging
            // blink out volts and tenths
            _delay_ms(100);
            uint8_t result = battcheck();
            blink(result >> 5, BLINK_SPEED/6);
            _delay_ms(BLINK_SPEED);
            blink(1,5);
            _delay_ms(BLINK_SPEED*3/2);
            blink(result & 0b00011111, BLINK_SPEED/6);
#else  // ifdef BATTCHECK_VpT
            // blink zero to five times to show voltage
            // (~0%, ~25%, ~50%, ~75%, ~100%, >100%)
            blink(battcheck(), BLINK_SPEED/6);
#endif  // ifdef BATTCHECK_VpT
            // wait between readouts
            _delay_s(); _delay_s();
        }
#endif // ifdef USE_BATTCHECK
        else if (output == GROUP_SELECT_MODE) {
            // exit this mode after one use
            mode_idx = 0;
//            mode_override = 0;

            for(i=0; i<NUM_MODEGROUPS; i++) {
                modegroup = i;
                save_state();

                blink(1, BLINK_SPEED/3);
            }
            _delay_s(); _delay_s();
        }
#ifdef TEMP_CAL_MODE
#ifdef TEMPERATURE_MON
        else if (output == TEMP_CAL_MODE) {
            // make sure we don't stay in this mode after button press
            mode_idx = 0;
//            mode_override = 0;

            // Allow the user to turn off thermal regulation if they want
            maxtemp = 255;
            save_state(); 
            set_mode(RAMP_SIZE/4);  // start somewhat dim during turn-off-regulation mode
            _delay_s(); _delay_s();

            // run at highest output level, to generate heat
            set_mode(RAMP_SIZE);

            // measure, save, wait...  repeat
            while(1) {
                maxtemp = get_temperature();
                save_state();
                _delay_s(); _delay_s();
            }
        }
#endif
#endif  // TEMP_CAL_MODE
        else {  // Regular non-hidden solid mode
            set_mode(actual_level);
#ifdef TEMPERATURE_MON  
			 
            uint8_t temp=get_temperature();

            // step down? (or step back up?)
            if (temp >= maxtemp) {
                overheat_count ++;
                // reduce noise, and limit the lowest step-down level
                if ((overheat_count > 8) && (actual_level > (RAMP_SIZE/8))) {
                    actual_level --;
                    //_delay_ms(5000);  // don't ramp down too fast
                    overheat_count = 0;  // don't ramp down too fast
                }
            } else {

                // if we're not overheated, ramp up to the user-requested level
                overheat_count = 0;
                if ((temp < maxtemp - 2) && (actual_level < output)) {
                    actual_level ++;
                }
            }
            set_mode(actual_level);
			#ifdef VOLTAGE_MON
			    ADC_on();  // switch back to voltage mode
			#endif

#endif
            // Otherwise, just sleep.
            _delay_ms(500);
            // If we got this far, the user has stopped fast-pressing.
            // So, don't enter config mode.
            //fast_presses = 0;
        }
        fast_presses = 0; //Theory here requires that every mode/strobe takes at least some time to get here.
		                  // see note in strobe() function.
#ifdef VOLTAGE_MON
        //if (ADCSRA & (1 << ADIF)) {  // if a voltage reading is ready
			                            //  FR says good idea, but it only takes 100us at worst.
										// let's save a few bytes of code instead.
            // See if voltage is lower than what we were looking for
            //uint8_t temp=get_voltage();
			//blink_value(temp);
            //if (temp < ADC_LOW) {
            if (get_voltage() < ADC_LOW) {
                lowbatt_cnt ++;
            } else {
                lowbatt_cnt = 0;
            }
            // See if it's been low for a while, and maybe step down
            if (lowbatt_cnt >= 8) {
                // DEBUG: blink on step-down:
                //set_level(0);  _delay_ms(100);

                if (actual_level > RAMP_SIZE) {  // hidden / blinky modes
                    // step down from blinky modes to medium
                    actual_level = RAMP_SIZE / 2;
                } else if (actual_level > 1) {  // regular solid mode
                    // step down from solid modes somewhat gradually
                    // drop by 25% each time
                    actual_level = (actual_level >> 2) + (actual_level >> 1);
                    // drop by 50% each time
                    //actual_level = (actual_level >> 1);
                } else { // Already at the lowest mode
                    // Turn off the light
                    set_level(0);
                    // Power down as many components as possible
                    set_sleep_mode(SLEEP_MODE_PWR_DOWN);
					cli(); // make sure not to wake back up.
					sleep_bod_disable();// power goes from 25uA to 4uA.
                    sleep_mode();
                }
                set_mode(actual_level);
                output = actual_level;
                lowbatt_cnt = 0;
                // Wait before lowering the level again
                //_delay_ms(250);
            }
            // Make sure conversion is running for next time through
            //ADCSRA |= (1 << ADSC);
        //}
#endif  // ifdef VOLTAGE_MON
    }// end of main while loop
    //return 0; // Standard Return Code
}


// The old way to do the toggles, in case we need it back:
            //// Enter or leave "muggle mode"?
            //toggle(&muggle_mode, 1);
            //if (muggle_mode) { continue; };  // don't offer other options in muggle mode
//
            //toggle(&memory, 2);
//
            //toggle(&enable_moon, 3);
//
            //toggle(&reverse_modes, 4);
//
            //// Enter the mode group selection mode?
            //mode_idx = GROUP_SELECT_MODE;
            //toggle(&mode_override, 5);
            //mode_idx = 0;
//
//#ifdef OFFTIM3
            //toggle(&offtim3, 6);
//#endif
//
//#ifdef TEMPERATURE_MON
            //// Enter temperature calibration mode?
            //mode_idx = TEMP_CAL_MODE;
            //toggle(&mode_override, 7);
            //mode_idx = 0;
//#endif

//#ifdef USE_FIRSTBOOT
            //toggle(&firstboot, 8);
//#endif
//#if defined(USE_ESWITCH) && (OTSM_PIN != ESWITCH_PIN)
            //toggle(&lockswitch, 9);
//#endif
